;; Generated package description from other-frame-window.el  -*- no-byte-compile: t -*-
(define-package "other-frame-window" "1.0.6.0.20201201.213818" "Minor mode to enable global prefix keys for other frame/window buffer placement" '((emacs "24.4")) :url "https://elpa.gnu.org/packages/other-frame-window.html" :keywords '("frame" "window") :authors '(("Stephen Leake" . "stephen_leake@member.fsf.org")) :maintainer '("Stephen Leake" . "stephen_leake@member.fsf.org"))
